unit SoftQuery;

interface

// =========================================================================
// An Faster Request Live Query
//
// Copyright 1999 Fernando Tolentino/Goyaz Tecnologia - Brazil
// All rights reserved.
//
// This components needs TKbmMemTable v. 1.22
// Copyright 1999 Kim Bo Madsen/Optical Services - Scandinavia
// All rights reserved.

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  kbmMemTable, Db, WSocket, CliProcess, Srv_Cli, Registry;

type
  TSoftQuery = class;

  TSoftConnection = class(TWSocket)
  public
    ConnectError: Integer;
    Buffer: TClientConnBuffer;
    constructor Create(AOwner:TComponent); override;
    destructor  Destroy; override;
    procedure CliSocketDataAvailable(Sender: TObject; Error: Word);
    procedure CliSocketSessionConnected(Sender: TObject; Error: Word);
    procedure CliSocketSessionClosed(Sender: TObject; Error: Word);
    procedure Connect; override;
    procedure Open( LoadRegistryParams, WriteRegistryParams:Boolean );
  end;

  TSoftQuery = class(TkbmCustomMemTable)
  private
    FTableName, FPrimaryKey, FEditFields, FIndexFieldNames: String;
    FSync, FManualMode, FLoading, FFieldDefsOk: Boolean;
    FSQL, FSQLInsert, FSQLDelete, FSQLUpdate: TStringList;
    procedure GenerateWriteSQL;
    procedure DestroyQuerys;
    procedure qPrepare(SQL: String; SQLCommand: TSQLCommand; FillParams:Boolean);
    function  qRun: Integer;
    procedure qDone;
    procedure SetIndexFieldNames( const Value: String );
    procedure SetTableName(const Value:string);
    procedure SetPrimaryKey(const Value:string);
    procedure SetEditFields(const Value:string);
    procedure SetSync(Value:Boolean);
    procedure SetManualMode(Value:Boolean);

    function  GetSQL:TStringList;
    procedure SetSQL(value:TStringList);
    function  GetSQLInsert:TStringList;
    procedure SetSQLInsert(value:TStringList);
    function  GetSQLDelete:TStringList;
    procedure SetSQLDelete(value:TStringList);
    function  GetSQLUpdate:TStringList;
    procedure SetSQLUpdate(value:TStringList);

  protected
    procedure InternalInitFieldDefs; override;
    procedure InternalOpen; override;
    procedure InternalClose; override;
    procedure InternalDelete; override;
    procedure InternalPost; override;
    procedure Sort;
  public
    constructor Create(AOwner:TComponent); override;
    destructor  Destroy; override;
    property Sync:Boolean read FSync write SetSync;
  published

    property TableName:String Read FTableName      write SetTableName;
    property PrimaryKey:String Read FPrimaryKey    write SetPrimaryKey;
    property EditFields:String Read FEditFields   write SetEditFields;
    property SQL:TStringList read GetSQL write SetSQL;
    property SQLInsert:TStringList read GetSQLInsert write SetSQLInsert stored FManualMode;
    property SQLDelete:TStringList read GetSQLDelete write SetSQLDelete stored FManualMode;
    property SQLUpdate:TStringList read GetSQLUpdate write SetSQLUpdate stored FManualMode;
    property ManualMode:Boolean read FManualMode write SetManualMode;
//    property IndexFieldNames: String read FIndexFieldNames write SetIndexFieldNames;
    property Active;
    property Filtered;
    property ReadOnly;
    property Filter;
    property MasterFields;
    property MasterSource;
    property Capacity;
    property OnLoadRecord;
    property OnLoadField;
    property OnCompressSave;
    property OnDecompressLoad;
    property OnCompressBlobStream;
    property OnDecompressBlobStream;
    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property OnCalcFields;
    property OnDeleteError;
    property OnEditError;
    property OnFilterRecord;
    property OnNewRecord;
    property OnPostError;
  end;

var
  SoftConnection: TSoftConnection=nil;

implementation

{ TSoftConnection }

constructor TSoftConnection.Create(AOwner: TComponent);
begin
  Inherited Create( AOwner);
  Buffer := TClientConnBuffer.Create( self );
  LineMode := False;
  LineEnd := #13#10;
  LineEcho := False;
  LineEdit := False;
  Proto := 'tcp';
  LocalAddr := '0.0.0.0';
  LocalPort := '0';
  MultiThreaded := False;
  OnDataAvailable := CliSocketDataAvailable;
  OnSessionClosed := CliSocketSessionClosed;
  OnSessionConnected := CliSocketSessionConnected;
  FlushTimeout := 60;
  SendFlags := wsSendNormal;
  LingerOnOff := wsLingerOn;
  LingerTimeout := 0;
  SocksLevel := '5';
  SocksAuthentication := socksNoAuthentication;

  Proto := 'tcp';
  Port  := 'telnet';
  Addr  := 'localhost';
end;

destructor  TSoftConnection.Destroy;
begin
  Buffer.Free;
  inherited Destroy;
end;

procedure   TSoftConnection.CliSocketDataAvailable(Sender: TObject; Error: Word);
begin
 Buffer.Parse;
end;

procedure   TSoftConnection.CliSocketSessionConnected(Sender: TObject;
  Error: Word);
begin
  ConnectError := Error;
  if Error <> 0 then
      Buffer.Error('Can''t connect, error #' + IntToStr(Error))
end;

procedure TSoftConnection.CliSocketSessionClosed(Sender: TObject; Error: Word);
begin
end;

procedure TSoftConnection.Connect;
begin
  Open( True, False)
end;

procedure TSoftConnection.Open;
var Key:String;
begin
  if LoadRegistryParams or WriteRegistryParams then begin
    Key := 'Software\Goyaz Tecnologia\'+Application.Title;
    with TRegistry.Create do try
      RootKey := HKEY_LOCAL_MACHINE;
      if not KeyExists(Key) then begin
        OpenKey(Key, True);
        WriteString('Protocol', Proto);
        WriteString('ServerAddress', Addr);
        WriteString('ServerPort', Port);
      end
      else OpenKeyReadOnly(Key);

      Self.Close;

      Proto := ReadString('Protocol');
      Addr  := ReadString('ServerAddress');
      Port  := ReadString('ServerPort');

    finally
      Free;
    end;
  end;
  Inherited Connect;
  while State in [wsConnecting] do ProcessMessage;
end;

constructor TSoftQuery.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  FSQL := TStringList.Create;
  FSQLInsert := TStringList.Create;
  FSQLDelete := TStringList.Create;
  FSQLUpdate := TStringList.Create;
  FSync := True;
  FManualMode := False;
  FLoading := False;
end;

destructor  TSoftQuery.Destroy;
begin
  DestroyQuerys;
  FSQL.Free;
  FSQLInsert.Free;
  FSQLDelete.Free;
  FSQLUpdate.Free;
  inherited Destroy;
end;

procedure TSoftQuery.DestroyQuerys;
begin
  FSync := False;
  qDone;
  if Active then EmptyTable;
end;

procedure TSoftQuery.GenerateWriteSQL;
function ComputePrimaryKey( OldValues, NullValues : Boolean ):String;
var S, S1:string;
    Sep:Boolean;
begin
  Sep := False;
  S := FPrimaryKey;
  Result := '';
  while S <> '' do begin
    if Sep then Result := Result + ' and '#13'   ';
    Sep := True;
    S1 := strDivide(S, ';');
    Result := Result + S1;
    if NullValues then      Result := Result + ' is Null '
    else if NullValues then Result := Result + '=:' + S1 + ' '
    else                    Result := Result + '=:Old_' + S1 + ' ';
  end
end;
function ComputeSQLDelete( NullValues : Boolean ):String;
begin
  Result := 'Delete from '+FTableName+' where '+ComputePrimaryKey( True, NullValues ); 
end;

// Delete from tmp where cdExame is Null
//   Sybase BUG
//
// insert into tmp (cdExame,
//                  nmExame,
//                  InstAtendim)
//      values     (:cdExame,
//                  :nmExame,
//                  :InstAtendim)
//
// select cdExame, nmExame, InstAtendim
//        from Exame
//        where nmExame=:nmExame
procedure GenerateInsert;
var _SQL, _Into, _Values, _Where, S, S1: String;
    Sep, IsBlob: Boolean;
    Field: TField;
begin
  _Into := '   ';
  _Values := '   ';
  _Where := '   ';
  _SQL := '';
{          '/* Sybase BUG - Delete na verdade nao faz nada */'#13+
          ComputeSQLDelete( True )  + #13#13; }

  Sep := False;
  S := FEditFields;
  while S <> '' do begin
    S1 := strDivide(S, ';');
    Field := FindField( S1 );
    if Field=nil then
      Raise Exception.Create('Invalid EditField='+S1);

    IsBlob := Field is TBlobField;

    if Sep then begin
      _Into := _Into + ','#13'   ';
      _Values := _Values + ','#13'   ';
      if not isBlob then
        _Where := _Where + ' and '#13'   ';
    end;

    Sep := True;
    _Into := _Into + S1;
    _Values := _Values + ':'+ S1;
    if not isBlob then
      _Where := _Where + S1 + '=:' + S1;
  end;

  _SQL := _SQL +  '/* Insere o registro */'#13+
                  'insert into '+ FTableName +' ('#13 +
                  _Into + #13 +
                  ')'#13 +
                  'values ('#13 +
                  _Values + #13 +
                  ')'#13 +
                  #13;

{  if FRefreshInsertion then begin
    _SQL := _SQL +  '/* Retorna o registro */'#13+
                    'select '#13 +
                    _Into + #13 +
                    'from '+FTableName + #13 +
                    'where'#13'   ';

    if FRefreshByFields then
      _SQL := _SQL + _Where
    else
      _SQL := _SQL + ComputePrimaryKey(False, False);

  end;
}

  FSQLInsert.Text := _SQL;
end;

procedure GenerateUpdate;
var _From, _Values, S, S1: String;
    Sep: Boolean;
    Field: TField;
begin

  _From := '   ';
  _Values := '   ';
  Sep := False;
  S := FEditFields;
  while S <> '' do begin
    S1 := strDivide(S, ';');
    Field := FindField( S1 );
    if Field=nil then
      Raise Exception.Create('Invalid EditField='+S1);

    if Sep then begin
      _Values := _Values + ','#13'   ';
      _From := _From +','#13'   ';
    end;

    Sep := True;
    _Values := _Values + S1 + '=:' + S1;
    _From := _From + S1;

  end;

// update dbu.tmp set
//   cdExame=1020202,
//   nmExame='s3sss',
//   InstAtendim=null
//   where
//   cdexame=1020202;

// select cdExame,
// nmExame,
// InstAtendim
// from dbu.tmp
// where cdexame=1020202

  FSQLUpdate.Text := '/* Altera o registro */'#13+
                     'update '+FTableName+ ' set '#13+
                     _Values +#13+
                     'where '+#13+
                     ComputePrimaryKey(True, False) +#13+
                     #13;
{  if FRefreshUpdate then begin
  FSQLUpdate.Text := FSQLUpdate.Text +
                     '/* retorna o registro alterado */'#13+
                     'select ' + _From +#13+
                     'from '+FTableName +#13+
                     'where '#13+
                     ComputePrimaryKey(False, False)
  end;
}
end;

procedure GenerateDelete;
begin
  FSQLDelete.text := '/* Exclui o registro */'#13+
                     ComputeSQLDelete( False ) ;
end;

begin
  GenerateInsert;
  GenerateUpdate;
  GenerateDelete;
end;

procedure TSoftQuery.qPrepare;
var I: Integer;
    S: String;
begin

  SoftConnection.Buffer.BeginUpdate( 'qPrepare: '+inttostr(byte(SQLCommand)) + ' ' +sql);
  
  case SQLCommand of
    cmdSQLExec      : SoftConnection.Buffer.ExecSQL( self, SQL, True );
    cmdSQLOpen      : SoftConnection.Buffer.OpenSQL( self, SQL, True );
    cmdSQLFieldDefs : SoftConnection.Buffer.UpdateFieldDefs( self, SQL, True );
  end;

  if FillParams then
  begin

    SQL := UpperCase (SQL);
    for I:=0 to FieldCount-1 do
    begin
      S := uppercase(Fields[I].FieldName);
      if pos(':'+S, SQL)>0 then begin
        SoftConnection.Buffer.AddSQLParam(Fields[I].FieldName, Fields[I].DataType, Fields[I].Value)
      end;
      if pos(':OLD_'+S, SQL)>0 then begin
        SoftConnection.Buffer.AddSQLParam('Old_'+Fields[I].FieldName, Fields[I].DataType, Fields[I].OldValue)
      end;
    end;

  end;
end;

function TSoftQuery.qRun;
begin
  Result := SoftConnection.Buffer.runSQL;
end;

procedure TSoftQuery.qDone;
begin
  SoftConnection.Buffer.EndUpdate;
end;

procedure TSoftQuery.SetIndexFieldNames( const Value: String );
begin
  FIndexFieldNames := Value;
  Sort;
end;

procedure TSoftQuery.Sort;
begin
{  if Active then begin
    if SortFields <> FIndexFieldNames then begin
      IndexFields := FIndexFieldNames;
      SortFields := FIndexFieldNames;
    end;
    _Sync := FSync;
    _Loading := FLoading;
    FSync := False;
    FLoading := True;
    try
      inherited Sort([mtcoCaseInsensitive]);
    finally
      FLoading := _Loading;
      FSync := _Sync;
    end;

  end;}
end;

{procedure TSoftQuery.SetDatabaseName(const Value:string);
begin
  CheckInactive;
  FDatabaseName := Value;
end;
}

procedure TSoftQuery.SetTableName(const Value:string);
begin
  CheckInactive;
  FTableName := Value;
end;

procedure TSoftQuery.SetPrimaryKey(const Value:string);
begin
  CheckInactive;
  FPrimaryKey := Value;
end;

procedure TSoftQuery.SetEditFields(const Value:string);
begin
  CheckInactive;
  FEditFields := Value;
end;

procedure TSoftQuery.SetSync(Value:Boolean);
begin
  CheckBrowseMode;
  if State=dsBrowse then FSync := Value;
end;

procedure TSoftQuery.SetManualMode(Value:Boolean);
begin
  CheckInactive;
  FManualMode := Value;
  if not (FManualMode) and Active then GenerateWriteSQL;
end;

function  TSoftQuery.GetSQL:TStringList;
begin
  Result := FSQL;
end;

procedure TSoftQuery.SetSQL(value:TStringList);
begin
  Close;
  DestroyQuerys;
  FSQL.Assign(value);
end;

function  TSoftQuery.GetSQLInsert:TStringList;
begin
  Result := FSQLInsert;
end;

procedure TSoftQuery.SetSQLInsert(value:TStringList);
begin
  FSQLInsert.Assign(value);
end;

function  TSoftQuery.GetSQLDelete:TStringList;
begin
  Result := FSQLDelete;
end;

procedure TSoftQuery.SetSQLDelete(value:TStringList);
begin
  FSQLDelete.Assign(value);
end;

function  TSoftQuery.GetSQLUpdate:TStringList;
begin
  Result := FSQLUpdate;
end;

procedure TSoftQuery.SetSQLUpdate(value:TStringList);
begin
  FSQLUpdate.Assign(value);
end;

procedure TSoftQuery.InternalInitFieldDefs;
begin
  if not FLoading then begin
    qPrepare( FSQL.Text, cmdSQLFieldDefs, True);
    try
      qRun;
    finally
      qDone;
    end;
  end;
end;

procedure TSoftQuery.InternalOpen;
var i:integer;
begin
  if FLoading then Inherited InternalOpen
  else begin
    if FTableName='' then
      Raise Exception.Create('TableName can''t be null');
    if FPrimaryKey='' then
      Raise Exception.Create('PrimaryKey can''t be null');

    FLoading := True;
    try
      FSync := False;

      qPrepare(FSQL.Text, cmdSQLOpen, True);
      try
        qRun; // inherited InternalOpen;

        SetIndexFieldNames(FIndexFieldNames);

        Sort;

        if not (FManualMode) and not(ReadOnly) then GenerateWriteSQL;
        
      finally
        qDone;
      end;
      FSync := True;
    finally
      FLoading := False;
    end;

    if FEditFields='' then begin
      for i:=0 to FieldDefs.Count-1 do begin
        if i>1 then FEditFields := FEditFields + ';';
        FEditFields := FEditFields + FieldDefs[I].Name
      end;
    end;
  end;
end;

procedure TSoftQuery.InternalClose;
begin
  DestroyQuerys;
  FFieldDefsOk := False;
  inherited InternalClose;
end;

procedure TSoftQuery.InternalPost;
var I: Integer;
begin
  if FSync then begin
    if State=dsInsert then begin
      qPrepare(FSQLInsert.text, cmdSQLExec, True);
      try
        qRun;
      finally
        qDone;
      end;
    end
    else begin
      qPrepare(FSQLUpdate.Text, cmdSQLExec, True);
      try
        I := qRun;
      finally
        qDone;
      end;
      if I <> 1 then
        raise Exception.Create('Update fail. RowsAffected='+inttostr(I));
    end;
  end;
  inherited InternalPost;
//  if not FLoading then Sort;
end;

procedure TSoftQuery.InternalDelete;
var I:Integer;
begin
  if FSync then begin
    qPrepare(FSQLDelete.Text, cmdSQLExec, True);
    try
      I:=qRun;
    finally
      qDone;
    end;
    if I <> 1 then
      raise Exception.Create('Delete fail. RowsAffected='+inttostr(I));
  end;
  inherited InternalDelete;
end;

initialization
  SoftConnection:=TSoftConnection.Create(nil)

finalization
  SoftConnection.Free;

end.

